from enum import Enum, unique


@unique
class Rotation(Enum):
    """
        Rotation
        An enumeration to indicate the rotation of a Layout.
    """
    UP = 0  # Normal
    RIGHT = 1  # rotate left 90°
    DOWN = 2  # rotate 180°
    LEFT = 3  # rotate right 90°



from PIL import Image, ImageDraw
import numpy
import unittest

from rpi_inky_layout import Layout, Rotation
from . import layout_fixtures as fixtures


class TestLayoutRotation(unittest.TestCase):

    @staticmethod
    def writeImage(layout, filename):
        layout.write('test/expected-images/' + filename)

    def testRotation(self):
        self.assertEqual(0, Rotation.UP.value)
        self.assertEqual(1, Rotation.RIGHT.value)
        self.assertEqual(3, Rotation.LEFT.value)
        self.assertEqual(2, Rotation.DOWN.value)

    def testCreateLayoutRotatedUp(self):
        layout = Layout((200, 100), rotation=Rotation.UP)
        self.assertEqual((200, 100), layout.size)
        self.assertEqual(0, layout.rotation_degrees)

    def testCreateLayoutRotatedLeft(self):
        layout = Layout((200, 100), rotation=Rotation.LEFT)
        self.assertEqual((100, 200), layout.size)
        self.assertEqual(270, layout.rotation_degrees)

    def testCreateLayoutRotatedDown(self):
        layout = Layout((200, 100), rotation=Rotation.DOWN)
        self.assertEqual((200, 100), layout.size)
        self.assertEqual(180, layout.rotation_degrees)

    def testCreateLayoutRotatedRight(self):
        layout = Layout((200, 100), rotation=Rotation.RIGHT)
        self.assertEqual((100, 200), layout.size)
        self.assertEqual(90, layout.rotation_degrees)

    def testLayoutRotatedRightAdd3Layers(self):
        self.add3LayersAndRotate((300, 100), Rotation.RIGHT)

    def testLayoutRotatedLeftAdd3Layers(self):
        self.add3LayersAndRotate((300, 100), Rotation.LEFT)

    def testLayoutRotatedDownAdd3Layers(self):
        self.add3LayersAndRotate((100, 300), Rotation.DOWN)

    def testLayoutRotatedUpAdd3Layers(self):
        self.add3LayersAndRotate((100, 300), Rotation.UP)

    def add3LayersAndRotate(self, size, rotation):
        layout = fixtures.threeLayers(
            size, packingMode='v', border=0, rotation=rotation
        )
        img = Image.new("RGB", layout.size, 0x000)
        layout.setImage(img)
        [
            self.assertEqual((100, 100), child.size)
            for child
            in layout.children
        ]

        layout1 = layout.children[0]
        layout2 = layout.children[1]
        layout3 = layout.children[2]
        self.assertEqual((0, 0), layout1.topLeft)
        self.assertEqual((0, 100), layout2.topLeft)
        self.assertEqual((0, 200), layout3.topLeft)
        img1 = Image.new("RGB", layout1.size, 0xff8800)  # red
        self.drawTextOnImage("layout1", img1)
        layout1.setImage(img1)
        img2 = Image.new("RGB", layout2.size, 0x000ff00)  # green
        self.drawTextOnImage("layout2", img2)
        layout2.setImage(img2)
        img3 = Image.new("RGB", layout3.size, 0xaaaaff)  # blue
        self.drawTextOnImage("layout3", img3)
        layout3.setImage(img3)
        self.writeImage(
            layout,
            "test-rotated-{rot}-add-3-layers.png".format(rot=rotation.name)
        )

    def drawTextOnImage(self, text, img):
        draw = ImageDraw.Draw(img)
        tsize = draw.textsize(text)
        tdiff = numpy.subtract(img.size, tsize)
        thalf = numpy.divide(tdiff, 2)
        tpos = tuple(thalf)

        draw.text(tpos, text, 0x000000)


if __name__ == '__main__':
    unittest.main()



from rpi_inky_layout import Layout, Rotation


# A file of pre-defined layout fixtures to re-use in tests.
def childlessLayout(size=(200, 100), border=0, packingMode='h'):
    return Layout(size, packingMode=packingMode, border=border)


def layoutWithLayers(
        size,
        count,
        border=0,
        packingMode='h',
        rotation=Rotation.UP
):
    layout = Layout(
        size,
        packingMode=packingMode,
        border=border,
        rotation=rotation
    )
    [layout.addLayer() for i in range(count)]
    return layout


def oneLayer(size=(97, 85), border=0, packingMode='h'):
    return layoutWithLayers(size, 1, border, packingMode=packingMode)


def twoLayers(size=(219, 100), border=0, packingMode='h'):
    return layoutWithLayers(size, 2, border, packingMode=packingMode)


def threeLayers(
        size=(200, 100),
        border=0,
        packingMode='h',
        rotation=Rotation.UP
):
    return layoutWithLayers(
        size, 3, border,
        packingMode=packingMode, rotation=rotation
    )


def fourLayers(size=(203, 100), border=0, packingMode='h'):
    return layoutWithLayers(size, 4, border, packingMode=packingMode)


def fiveLayers(size=(197, 100), border=0, packingMode='h'):
    return layoutWithLayers(size, 5, border, packingMode=packingMode)


def sixLayers(size=(197, 100), border=0, packingMode='h'):
    return layoutWithLayers(size, 6, border, packingMode=packingMode)



class Position:

    @staticmethod
    def average(_tuple):
        return (_tuple[0] - _tuple[1])/2

    @staticmethod
    def middle_centre(box_wh, object_wh):
        sizes = zip(box_wh, object_wh)
        return tuple([Position.average(size) for size in sizes])

    @staticmethod
    def middle_left(box_wh, object_wh):
        box_h = box_wh[1]
        object_h = object_wh[1]
        return (0, Position.average((box_h, object_h)))

    @staticmethod
    def middle_right(box_wh, object_wh):
        box_w, box_h = box_wh
        object_w, object_h = object_wh
        return (box_w - object_w, Position.average((box_h, object_h)))

    @staticmethod
    def top_left(box_wh, object_wh):
        return (0, 0)

    @staticmethod
    def top_centre(box_wh, object_wh):
        box_w = box_wh[0]
        object_w = object_wh[0]
        return (Position.average((box_w, object_w)), 0)

    @staticmethod
    def top_right(box_wh, object_wh):
        box_w, box_h = box_wh
        object_w, object_h = object_wh
        return (box_h - object_h, 0)

    @staticmethod
    def bottom_left(box_wh, object_wh):
        box_h = box_wh[1]
        object_h = object_wh[1]
        return (0, box_h - object_h)

    @staticmethod
    def bottom_centre(box_wh, object_wh):
        box_w, box_h = box_wh
        object_w, object_h = object_wh
        return (Position.average((box_w, object_w)), box_h - object_h)

    @staticmethod
    def bottom_right(box_wh, object_wh):
        b_w, b_h = box_wh
        o_w, o_h = object_wh
        return (b_w - o_w, b_h - o_h)



from PIL import Image, ImageDraw
from random import randint
import unittest

from rpi_inky_layout import Layout
from . import layout_fixtures as fixtures


class TestLayoutVerticalBasics(unittest.TestCase):

    @staticmethod
    def drawAndWrite(layout, filename):
        layout.write('test/expected-images/' + filename)

    def testAdd2ChildrenVerticalPacking(self):
        layout = fixtures.oneLayer((200, 100), packingMode='v')
        layer1 = layout.children[0]
        # VERIFIED
        self.assertEqual((200, 100), layer1.size)
        layer2 = layout.addLayer()
        self.assertEqual((200, 50), layer1.size)
        self.assertEqual((200, 50), layer2.size)
        layer1.setImage(self.NewImage(layer1, "RGB"))
        self.addStuffToImage(layer1, 1, self.RandomColour())
        layer2.setImage(self.NewImage(layer2, "RGB"))
        self.addStuffToImage(layer2, 1, self.RandomColour())

        self.drawAndWrite(layout, 'testAdd2ChildrenVerticalPacking.png')

    # REVISIT
    def testAddChildrenPackedVWithBorderEnabledOneLayers(self):
        layout = fixtures.oneLayer((200, 300), packingMode='v', border=1)
        [
            self.assertEqual((198, 298), child.size)
            for child
            in layout.children

        ]

    def testAddChildrenPackedVWithBorderEnabledTwoLayers(self):
        layout = fixtures.twoLayers((200, 300), packingMode='v', border=1)
        [
            self.assertEqual((198, 148), child.size)
            for child
            in layout.children

        ]

    def testAddChildrenPackedVWithBorderEnabledThreeLayers(self):
        layout = fixtures.threeLayers((200, 300), packingMode='v', border=1)
        [
            self.assertEqual((198, 98), child.size)
            for child
            in layout.children

        ]

    def testAddChildrenPackedVWithBorderEnabledFourLayers(self):
        layout = fixtures.fourLayers((200, 300), packingMode='v', border=1)
        [
            self.assertEqual((198, 73), child.size)
            for child
            in layout.children

        ]

    def testAddChildrenPackedVWithBorderEnableFiveLayers(self):
        layout = fixtures.fiveLayers((200, 300), packingMode='v', border=1)
        [
            self.assertEqual((198, 58), child.size)
            for child
            in layout.children

        ]
        [
            self.addStuffToImage(child, 1, self.RandomColour())
            for child
            in layout.children
        ]
        self.drawAndWrite(layout, "addChildrenPackedVWiBorder5Layers.png")

    # FIXME refactor into separate utility package
    def NewImage(self, layout, mode='P', colour=randint(0, 2)):
        img = Image.new(mode, layout.size, colour)
        if mode == 'P':
            img.putpalette(Layout.DEFAULT_PALETTE)
        return img

    # FIXME refactor into separate utility package
    def RandomColour(self):
        return (randint(64, 255), randint(64, 255), randint(64, 255))

    # FIXME refactor into separate utility package
    def NewDraw(self, img):
        return ImageDraw.Draw(img)

    # FIXME refactor into separate utility package
    def addStuffToImage(self, layout, childId=0, bgcolour=(128, 128, 1)):

        img = self.NewImage(layout, "RGB", bgcolour)
        draw = self.NewDraw(img)
        text = "{id}\nchildId:{cid}\ndepth:{d}\ns:{size}\ntl:{tl}".format(
            id=layout._id,
            cid=childId, d=layout._depth, size=layout.size, tl=layout.topLeft)
        tsize = draw.textsize(text)
        if(tsize[1] > layout.size[1]):
            text = "{id}/{cid} d:{d}\ns:{size} tl:{tl}".format(
                id=layout._id, cid=childId, d=layout._depth,
                size=layout.size, tl=layout.topLeft)
        # draw.rectangle((0, 0), 1)
        draw.text((5, 5), text)
        layout.setImage(img)
        layout.draw()

    # FIXME refactor into separate utility package
    def InterrogateLayer(self, layout):
        print("parent size: {ps}".format(ps=layout.size))
        print("parent:borders: {b}".format(b=layout.borders))
        print("parent:drawable-size: {s}".format(s=layout._drawableSize()))
        print("parent:slot-width: {s}".format(s=layout._slotWidth()))
        for i, child in enumerate(layout._children):

            print("child-{i}:top-left: {tl}".format(i=i, tl=child.topLeft))
            print("child-{i}:size: {s}".format(i=i, s=child.size))


if __name__ == '__main__':
    unittest.main()



import unittest
from rpi_inky_layout import Layout, Rotation


class TestIssue37(unittest.TestCase):

    def testIssue37(self):
        i_res = (250, 122)
        layout = Layout(i_res, packingMode='v', rotation=Rotation.LEFT)
        _ratesLayer = layout.addLayer()
        _time = layout.addLayer()
        _attrs = _ratesLayer.addLayer(packingBias=1)
        _rates = _ratesLayer.addLayer(packingBias=3)

        self.assertEqual((30, 125), _attrs.size)
        self.assertEqual((90, 125), _rates.size)
        self.assertEqual((122, 125), _time.size)
        self.assertEqual(0, _ratesLayer._showSparePixels())


if __name__ == '__main__':
    unittest.main()



import unittest
from rpi_inky_layout import IndexOrder


class TestLayoutIndexCount(unittest.TestCase):

    def testIndexCount(self):
        self.assertEqual([0], IndexOrder.alternating(1))
        self.assertEqual([0, 1], IndexOrder.alternating(2))
        self.assertEqual([0, 2, 1], IndexOrder.alternating(3))
        self.assertEqual([0, 3, 1, 2], IndexOrder.alternating(4))
        self.assertEqual([0, 4, 1, 3, 2], IndexOrder.alternating(5))
        self.assertEqual([0, 5, 1, 4, 2, 3], IndexOrder.alternating(6))
        self.assertEqual([0, 6, 1, 5, 2, 4, 3], IndexOrder.alternating(7))


if __name__ == '__main__':
    unittest.main()




import unittest
from rpi_inky_layout import Position


class TestLayoutRotation(unittest.TestCase):

    def testTopLeft(self):
        self.assertEqual((0, 0), Position.top_left((10, 10), (8, 8)))

    def testTopCentre(self):
        self.assertEqual((1, 0), Position.top_centre((10, 10), (8, 8)))

    def testTopRight(self):
        self.assertEqual((2, 0), Position.top_right((10, 10), (8, 8)))

    def testMiddleLeft(self):
        self.assertEqual((0, 1), Position.middle_left((10, 10), (8, 8)))

    def testMiddleCentre(self):
        self.assertEqual((1, 1), Position.middle_centre((10, 10), (8, 8)))

    def testMiddleRight(self):
        self.assertEqual((2, 1), Position.middle_right((10, 10), (8, 8)))

    def testBottomLeft(self):
        self.assertEqual((0, 2), Position.bottom_left((10, 10), (8, 8)))

    def testBottomCentre(self):
        self.assertEqual((1, 2), Position.bottom_centre((10, 10), (8, 8)))

    def testBottomRight(self):
        self.assertEqual((2, 2), Position.bottom_right((10, 10), (8, 8)))



#!/usr/bin/env python3
from PIL import Image, ImageDraw
import unittest

from rpi_inky_layout import Layout


class TestLayoutAlternatePackingMode(unittest.TestCase):

    @staticmethod
    def writeImage(layout, filename):
        layout.write('test/expected-images/' + filename)

    def test_getChildSlotTotalAndBiasStart(self):
        layout = Layout((100, 100))
        self.assertEqual(0, layout._getChildSlotTotal())

        layout1 = layout.addLayer(packingMode='h')
        self.assertEqual(1, layout._getChildSlotTotal())
        self.assertEqual((100, 100), layout.size)
        self.assertEqual((100, 100), layout1.size)

        layout11 = layout1.addLayer()
        layout12 = layout1.addLayer()
        self.assertEqual((100, 100), layout.size)
        self.assertEqual((100, 100), layout1.size)
        self.assertEqual((50, 100), layout11.size)
        self.assertEqual((50, 100), layout12.size)

        layout2 = layout.addLayer(packingMode='v')
        self.assertEqual((100, 100), layout.size)
        self.assertEqual((50, 100), layout1.size)
        self.assertEqual((50, 100), layout2.size)
        self.assertEqual((25, 100), layout11.size)
        self.assertEqual((25, 100), layout12.size)

        layout21 = layout2.addLayer()
        layout22 = layout2.addLayer()

        self.assertEqual((100, 100), layout.size)
        self.assertEqual((50, 100), layout1.size)
        self.assertEqual((25, 100), layout11.size)
        self.assertEqual((25, 100), layout12.size)
        self.assertEqual((50, 100), layout2.size)
        self.assertEqual((50, 50), layout21.size)
        self.assertEqual((50, 50), layout22.size)
        layouts = (layout11, layout12, layout21, layout22)
        [
            self._setImage(layout, (0x40*index, 0xff, 0xff), str(layout._id))
            for index, layout in enumerate(layouts)
        ]
        self.writeImage(layout, "test-alternatePackingMode.png")

    def _setImage(self, layout, colour, text):
        img = Image.new("HSV", layout.size, colour)
        draw = ImageDraw.Draw(img)
        draw.text((0, 0), text)
        layout.setImage(img)


if __name__ == '__main__':
    unittest.main()



import unittest
from .layout_fixtures import childlessLayout


class TestLayoutNoLayers(unittest.TestCase):

    def setUp(self):
        self.layout = childlessLayout()

    def testChildCount(self):
        self.assertEqual(0, self.layout._childCount())

    def testHasOneChild(self):
        self.assertFalse(self.layout._onlyOneChild())

    def testHasMoreThanOneChild(self):
        self.assertFalse(self.layout._moreThanOneChild())

    def testHasChildren(self):
        self.assertFalse(self.layout._hasChildren())



__name__ = 'rpi_inky_layout'
from .layout import Layout  # noqa: F401
from .rotation import Rotation  # noqa: F401
from .position import Position  # noqa: F401
from .index_order import IndexOrder  # noqa: F401



from PIL import Image, ImageDraw
from random import randint
import numpy
from numpy import floor

from .rotation import Rotation
from .index_order import IndexOrder


class Layout:
    """
        Layout - a simple image layout manager for RPi Inky HATs.
    """

    DEFAULT_PALETTE = (255, 255, 255, 0, 0, 0, 255, 0, 0) + (0, 0, 0) * 252

    @staticmethod
    def tupleReversed(tupleValue):
        return tuple(reversed(tupleValue))

    def __str__(self):
        return "{id}/{depth}/{pm}/{pb}/{children}:{tl}+{dim}".format(
            id=self._id,
            depth=self._depth,
            tl=self.topLeft,
            dim=self.size,
            pb=self.packingBias,
            pm=self.packingMode,
            children=self._childCount()
        ) + "\nsl:{sl},b:{b},padd:{ps},sp:{sp},tl:{tl}".format(
                sl=self._slots,
                sp=self._spacers,
                b=self.borders,
                ps=self._paddings,
                tl=self._topLefts
        )

    def __init__(
        self, size=(250, 122), packingMode='h', border=(0, 0),
        depth=0, rotation=Rotation.UP, packingBias=1, imageMode="RGB"
    ):
        """
            Construct a new Layer.

            Parameters
            ----------
            size: tuple
                A 2-tuple describing the size of the layer.
                Default:(250,122), the size of an Inky-PHAT.
            packingMode: str
                The packing mode: one of 'h' or 'v'.
            border: int|(0,0)
                If an int, then the width, and the colour is assumed to be '2'.
                If a 2-tuple, then the first part is the width, and the second
                is the colour.
            depth: int
                for tracking the tree hierarchy - for internal use
                only. Default:0.
            rotation: Rotation
                The type of rotation. This modifies the way that
                images are drawn, so that UP becomes the default top, DOWN
                would render upside down, and left and right similarly modify
                the rendering.
            packingBias: int
                The packing bias. If you add 2 layers: one with
                default bias, one with packingBias=3, then the second layer
                will take up 3/4 of the width (3+1 = 4 = 100%).
            imageMode: str
                The image mode to use when creating a default image (when one
                does not exist). Refer to
                [Pillow image modes](https://pillow.readthedocs.io/en/latest/handbook/concepts.html#modes).
        """  # noqa: E501
        self.rotation = rotation
        self.packingBias = packingBias
        self.rotation_degrees = self.rotation.value * 90
        self.topLeft = (0, 0)
        self.drawBorders = True
        self.packingMode = packingMode
        self.size = size
        if self.rotation.value % 2:
            self.size = self.tupleReversed(self.size)

        self.imageMode = imageMode

        if isinstance(border, int):
            self.borders = (border, border, border, border)
            self.borderColour = 2
        elif isinstance(border, tuple):
            borders, self.borderColour = border
            if isinstance(borders, int):
                self.borders = (borders, borders, borders, borders)
            elif len(borders) == 2:
                btb, blr = borders
                self.borders = (btb, blr, btb, blr)
            elif len(borders) == 4:
                self.borders = borders
            else:
                raise("Illegal borders format: {b}".format(b=borders))

        elif not isinstance(border, None):
            raise("borders is not an allowed type:{t}".format(t=type(border)))

        self.children = []
        self._spacers = []
        self._paddings = []
        self._topLefts = []
        self._slots = []
        self._slotSizeError = 0.0
        self._image = None
        self._depth = depth
        self._id = randint(1024, 8192)

    def transformAsNeeded(self, twod):
        if self.packingMode == 'v':
            twod = self.tupleReversed(twod)
        return twod

    def setImage(self, image):
        """
            Set the image on this layer after you've drawn it.
            The image you set will be cropped to the Layer's size before being
            set on it.
        """
        self._image = image.crop((0, 0) + self.size)
        return self._image

    def addLayer(
            self,
            border=0,
            packingBias=1,
            packingMode='h',
            rotation=Rotation.UP
    ):
        """
            Add a new child layout to this layout as a layer.

            Parameters
            ------
            border: int | (int, colour)
                the same as the constructor's border property.
            packingBias: int
                the same as the constructor's packingBias property.
        """

        childLayer = Layout(
            depth=self._depth + 1,
            border=border,
            packingBias=packingBias,
            packingMode=packingMode,
            rotation=rotation
            )
        self.children.append(childLayer)
        self._resizeChildren()
        return childLayer

    def resize(self, size):
        self.size = size
        if self._image:
            self._image = self._image.crop((0, 0) + size)
        self._resizeChildren()

    def _childCount(self):
        return len(self.children)

    def _hasChildren(self):
        return self._childCount() > 0

    def _onlyOneChild(self):
        return self._childCount() == 1

    def _moreThanOneChild(self):
        return self._childCount() > 1

    def _calcDrawableSize(self):
        w, h = self.size
        bt, br, bb, bl = self.borders
        # Remove the borders
        w = w - bl - br
        h = h - bt - bb
        _calcDrawableSize = (w, h)
        return _calcDrawableSize

    def _calcSlotSize(self):
        dw, dw1 = self.transformAsNeeded(self._calcDrawableSize())
        sw = self._sumSpacersWidth()
        w = dw - sw
        if self._onlyOneChild():
            return (w, dw1)
        w = w * 1.0
        slotCount = self._getChildSlotTotal()
        if slotCount > 1:
            w = w / slotCount
        intW = int(floor(w))  # FIXME this underuses the drawable size
        self._slotSizeError = (w - intW) * slotCount
        print("SLOT SIZE ERROR:", self._slotSizeError)
        slotSize = (intW, dw1)
        return slotSize

    def _getSlotSizeFor(self, child):
        """get the slot size for this child.
        Accounts for the packingBias in the slot size.
        """
        slotSize = self._slotSize
        if self._moreThanOneChild():
            slotSize = (slotSize[0] * child.packingBias, slotSize[1])
        return self.transformAsNeeded(slotSize)

    def _calcPaddings(self):
        """The additional width to put into each spacer, to even things out."""
        drawableWidth, dh = self.transformAsNeeded(self._calcDrawableSize())
        drawableWidth = float(drawableWidth)
        _errorWidth = int(floor(self._slotSizeError + 0.49))
        print("_calcPaddings: _error:", _errorWidth)
        _paddingCount = self._childCount() - 1
        _paddings = [0] * _paddingCount
        _indexes = IndexOrder.alternating(_paddingCount)
        while _errorWidth > 0:
            _errorWidth -= 1
            _index = _errorWidth % _paddingCount
            _paddings[_indexes[_index]] += 1
        return _paddings

    def _calcIdealSpacerWidth(self):
        bt, br, bb, bl = self.borders
        if self.packingMode == 'h':
            return bl
        else:  # self.packingMode == 'v'
            return bt

    def _calcRealSpacerWidth(self, index):
        _optWidth = self._calcIdealSpacerWidth()
        _padding = self._paddings[index]
        return _optWidth + _padding

    def _calcTopLeft(self, index):
        slotStart = 0
        if index > 0:
            slotStart = self._getChildSlotStart(index)  # all previous ones
        slotDelta = slotStart * self._slotSize[0]
        topLeftBorders = (self.borders[0], self.borders[3])
        spacer = 0  # 0 offset from drawable area, NOT parent area
        if (index > 0):
            spacer = sum(self._spacers[0:index])
        top = spacer + slotDelta + topLeftBorders[0]
        left = topLeftBorders[1]
        topLeft = (top, left)
        return self.transformAsNeeded(topLeft)

    def _sumSpacersWidth(self):
        """The total width of all the spacers."""
        return sum(self._spacers)

    def _showSparePixels(self):
        if self.packingMode == 'h':
            borderIndex = 1
        else:
            borderIndex = 0
        size = self.transformAsNeeded(self.size)
        size0 = size[0]
        sumSpacers = sum(self._spacers)
        slotDims = [self.transformAsNeeded(slot)[0] for slot in self._slots]
        sumSlots = sum(slotDims)

        border1 = self.borders[1 - borderIndex]
        border2 = self.borders[3 - borderIndex]
        _sparePixels = (
            size0 - sumSpacers -
            sumSlots -
            border1 - border2
        )
        if _sparePixels > 0:
            print(
                "!!! SPARE PIXELS: {sp} on {s}".format(
                    sp=_sparePixels, s=self._spacers
                )
            )
        return _sparePixels

    def _resizeChildren(self):
        childCount = self._childCount()

        if childCount > 0:  # only update when there's children present

            # spacers - first pass
            self._spacers = [
                    self._calcIdealSpacerWidth() for c in
                    range(self._childCount() - 1)
            ]

            # calculate all the slots - slotSize needed for slots and topLefts.
            self._slotSize = self._calcSlotSize()
            self._slots = [
                self._getSlotSizeFor(child)
                for child
                in self.children
            ]
            # outcome: there is slot error!
            print("_resizeChildren: slotError", self._slotSizeError)
            # next, calculate padding
            self._paddings = self._calcPaddings()
            # add padding to existing spacers
            __spacers = [
                x[0] + x[1]
                for x
                in zip(self._spacers, self._paddings)
            ]
            # and reset the spacers with adjusted values
            self._spacers = __spacers
            # finally, toplefts depend on slots and adjusted spacers
            self._topLefts = [
                self._calcTopLeft(i) for i in range(self._childCount())
            ]

            self._showSparePixels()
            [
                self._resizeAChild(
                    child,
                    index
                ) for index, child in enumerate(self.children)
            ]

    def _resizeAChild(self, child, index):

        slotSize = self._slots[index]
        topLeft = self._topLefts[index]

        child.resize(slotSize)
        child.topLeft = topLeft

    def _getChildSlotTotal(self):
        """The total number of slots."""
        return self._getChildSlotCount(self.children)

    @staticmethod
    def _getChildSlotCount(childrenList):
        """The count of the slot."""
        packingBias = [child.packingBias for child in childrenList]
        packingCount = sum(packingBias)
        return packingCount

    def _getChildSlotStart(self, index):
        """Which slot this child starts in."""
        return self._getChildSlotCount(self.children[0:index])

    def draw(self):
        if not self._image:
            self._image = Image.new(self.imageMode, self.size, self._depth)
        self._drawChildren()

        # draw the border
        draw = ImageDraw.Draw(self._image)
        if self.drawBorders:
            self._drawBorder(draw)

        # return image rotated to the correct orientation
        return self._image.rotate(-self.rotation_degrees, expand=1)

    def _drawBorder(self, draw):
        w = self.borders[0]
        c = self.borderColour
        while w > 0:
            rectSize = tuple(numpy.subtract(self.size, w))
            w = w - 1
            draw.rectangle([(w, w), rectSize], outline=c, width=1)

        def drawSpacer(w, tl):
            if w > 0:
                if self.packingMode == 'h':
                    tlxy = [tl[0] - w, tl[1]]
                    brxy = [tl[0] - 1, self.size[1] - tl[1]]
                else:
                    tlxy = [tl[0], tl[1] - w]
                    brxy = [self.size[0] - tl[0], tl[1] - 1]
                tlbr = tlxy + brxy
                draw.rectangle(tlbr, fill=c, width=1)

        if len(self._spacers) > 0:
            [
                drawSpacer(w, tl)
                for w, tl
                in zip([0] + self._spacers, self._topLefts)
            ]

    def _drawChildren(self):
        [child._drawChildren() for child in self.children]
        [
            self._drawChildOnParent(child, index)
            for index, child
            in enumerate(self.children)
        ]

    def _drawChildOnParent(self, child, index):
        if not child._image:
            print("WARNING: NO IMAGE ON THIS CHILD", child)
        else:
            if not self._image:
                self._image = self.setImage(child._image.copy())
            else:
                self._image.paste(child._image, child.topLeft)

    def write(self, fp):
        self.draw().convert(mode="RGB").save(fp)
        if self.children:
            [
                child.write(fp.replace(".", "-{i}.".format(i=index)))
                for index, child in enumerate(self.children)
            ]



#!/usr/bin/env python3
from PIL import Image, ImageDraw
import unittest

from rpi_inky_layout import Layout


class TestLayoutPackingBias(unittest.TestCase):

    @staticmethod
    def writeImage(layout, filename):
        layout.write('test/expected-images/' + filename)

    def test_getChildSlotTotalAndBiasStart(self):
        layout = Layout((100, 100))
        self.assertEqual(0, layout._getChildSlotTotal())

        layout.addLayer(packingBias=3)
        self.assertEqual(3, layout._getChildSlotTotal())

        layout.addLayer(packingBias=3)
        self.assertEqual(6, layout._getChildSlotTotal())

        self.assertEqual(0, layout._getChildSlotStart(0))
        self.assertEqual(3, layout._getChildSlotStart(1))

    def testPackingBiasHorizontal3_1(self):
        layout = Layout((200, 10), packingMode='h', border=(0, 0))
        layout1 = layout.addLayer(packingBias=3)
        self.assertEqual((200, 10), layout1.size)

        layout2 = layout.addLayer(packingBias=1)

        self.assertEqual((150, 10), layout1.size)
        self.assertEqual((50, 10), layout2.size)
        self.assertEqual((0, 0), layout1.topLeft)
        self.assertEqual((150, 0), layout2.topLeft)

        self.setImage(layout1, 0xff0000, "layout1")
        self.setImage(layout2, 0x0088ff, "layout2")
        self.setImage(layout, 0x000000, "layout")
        self.writeImage(layout, "testPackingBiasHorizontal3_1.png")

    def testPackingBiasHorizontal1_3(self):
        layout = Layout((200, 10), packingMode='h', border=(0, 0))
        layout1 = layout.addLayer(packingBias=1)
        self.assertEqual((200, 10), layout1.size)

        layout2 = layout.addLayer(packingBias=3)

        print(layout)
        print(layout1)
        print(layout2)

        self.assertEqual((50, 10), layout1.size)
        self.assertEqual((150, 10), layout2.size)
        self.assertEqual((0, 0), layout1.topLeft)
        self.assertEqual((50, 0), layout2.topLeft)

        self.setImage(layout1, 0xff0000, "layout1")
        self.setImage(layout2, 0x0088ff, "layout2")
        self.setImage(layout, 0x000000, "layout")
        self.writeImage(layout, "testPackingBiasHorizontal1_3.png")

    def testPackingBiasVertical3_1(self):

        layout = Layout((10, 200), packingMode='v', border=(0, 0))
        layout1 = layout.addLayer(packingBias=3)
        self.assertEqual((10, 200), layout1.size)

        layout2 = layout.addLayer(packingBias=1)

        self.assertEqual((10, 150), layout1.size)
        self.assertEqual((10, 50), layout2.size)
        self.assertEqual((0, 0), layout1.topLeft)
        self.assertEqual((0, 150), layout2.topLeft)

        self.setImage(layout1, 0xff0000, "1")
        self.setImage(layout2, 0x0088ff, "2")
        self.setImage(layout, 0x000000, "L")
        self.writeImage(layout, "testPackingBiasVertical3_1.png")

    def testPackingBiasVertical1_3(self):

        layout = Layout((10, 200), packingMode='v', border=(0, 0))
        layout1 = layout.addLayer(packingBias=1)
        self.assertEqual((10, 200), layout1.size)

        layout2 = layout.addLayer(packingBias=3)

        self.assertEqual((10, 50), layout1.size)
        self.assertEqual((10, 150), layout2.size)
        self.assertEqual((0, 0), layout1.topLeft)
        self.assertEqual((0, 50), layout2.topLeft)
        self.setImage(layout1, 0xff0000, "1")
        self.setImage(layout2, 0x0088ff, "2")
        self.setImage(layout, 0x000000, "L")
        self.writeImage(layout, "testPackingBiasVertical1_3.png")

    def testDeepPackingBias(self):
        layout = Layout((200, 300), packingMode='v')
        layout1 = layout.addLayer(packingMode='h')
        layout2 = layout.addLayer()
        layout11 = layout1.addLayer(packingBias=1)
        layout12 = layout1.addLayer(packingBias=3)
        self.assertEqual((200, 150), layout2.size)
        self.setImage(layout11, 0xff0000, "Layout 11")
        self.setImage(layout12, 0x0088ff, "Layout 12")
        self.setImage(layout2, 0xaa0088, "Layout 2")
        self.writeImage(layout, "testDeepPackingBias.png")

    def testWrite(self):
        layout = Layout((200, 200))
        layout1 = layout.addLayer(packingBias=1)
        layout2 = layout.addLayer(packingBias=3)

        self.setImage(layout1, 0xff0000, "1")
        self.setImage(layout2, 0x0088ff, "2")
        self.setImage(layout, 0x000000, "L")
        self.writeImage(layout, "test-packingBias.png")

    @staticmethod
    def setImage(layout, colour, text):
        img = Image.new("RGB", layout.size, colour)
        draw = ImageDraw.Draw(img)
        draw.text((0, 0), text)
        layout.setImage(img)


if __name__ == '__main__':
    unittest.main()



class IndexOrder():

    @classmethod
    def alternating(cls, count):
        """"Generate an list containing alternate indexes.
            For distributing adjustments.
        """
        if count < 1:
            return []
        if count == 1:
            return [0]
        _lowIndexes = [n for n in range(0, int(count / 2))]
        _highIndexes = [n + count - 1 for n in range(0, int(-count / 2), -1)]
        _indexes = [0] * (len(_lowIndexes)+len(_highIndexes))
        _indexes[::2] = _lowIndexes
        _indexes[1::2] = _highIndexes
        if count % 2 == 1:
            _indexes = _indexes + [int(count / 2)]
        return _indexes



from os import path
import setuptools

VERSION = "v0.2.3"
DIST_FMT = "https://github.com/DoctorU/rpi-inky-layout/dist/{VERSION}.tar.gz"
DIST_URL = DIST_FMT.format(VERSION=VERSION)

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()
setuptools.setup(
  name='rpi-inky-layout',
  version='0.2.3',
  license='MIT',
  description="A layout companion library for Pimoroni's Inky HATs",
  long_description=long_description,
  long_description_content_type='text/markdown',
  author='Stephen Hobdell',
  author_email='doctoruseful@gmail.com',
  url='https://github.com/DoctorU/rpi-inky-layout',
  download_url=DIST_URL,
  packages=setuptools.find_packages(),
  keywords=['Layout', 'Framework', 'Image', 'Inky', 'Pimoroni'],
  install_requires=[
    'numpy>=1.19.5',
    'Pillow>=8.1.0'
  ],
  classifiers=[
    'Development Status :: 3 - Alpha',
    'Intended Audience :: Developers',
    'Topic :: Software Development :: Libraries :: Application Frameworks',
    'Topic :: Software Development :: Libraries :: Python Modules',
    'Topic :: Multimedia :: Graphics :: Presentation',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
    'Programming Language :: Python :: 3.9',
    "License :: OSI Approved :: MIT License",
    "Operating System :: Other OS"
  ]
)



from PIL import Image, ImageDraw
from random import randint
import unittest
from . import layout_fixtures as fixtures

from rpi_inky_layout import Layout


class TestLayoutBasics(unittest.TestCase):

    @staticmethod
    def drawAndWrite(layout, filename):
        layout.write('test/expected-images/' + filename)

    def testCreateLonelyLayout(self):
        layout = fixtures.childlessLayout()
        self.assertEqual((200, 100), layout.size)

        img = self.NewImage(layout)
        draw = ImageDraw.Draw(img)
        draw.text((2, 50), 'testCreateLayout')
        layout.setImage(img)
        self.drawAndWrite(layout, 'testCreateLayout.png')

    def testAdd2ChildrenHorizontalPacking(self):
        layout = Layout((200, 100), 'h', border=2)
        img = self.NewImage(layout)
        draw = self.NewDraw(img)
        draw.text((2, 50), "main layout 1", 2)
        layout.setImage(img)
        layer1 = layout.addLayer()

        img = self.NewImage(layer1)
        draw = self.NewDraw(img)
        draw.text((2, 50), "layer 1-1", 2)
        layer1.setImage(img)

        self.assertEqual((196, 96), layer1.size)
        self.drawAndWrite(layer1, 'Add2ChildrenHorizontalPacking1.png')

        layer2 = layout.addLayer()
        img = self.NewImage(layer2)
        draw = self.NewDraw(img)
        draw.text((2, 50), "layer 1-2", 2)
        layer2.setImage(img)

        self.assertEqual((97, 96), layer1.size)
        self.assertEqual((97, 96), layer2.size)
        self.drawAndWrite(layer2, 'Add2ChildrenHorizontalPacking2.png')
        self.drawAndWrite(layout, 'Add2ChildrenHorizontalPacking.png')

    def testAddChildrenPackedHWithBorderEnabledOneLayer(self):
        layout = fixtures.layoutWithLayers((600, 100), 1, border=1)
        [
            self.addStuffToImage(child, 1, self.RandomColour())
            for child
            in layout.children
        ]
        [
            self.assertEqual((598, 98), child.size)
            for child
            in layout.children
        ]

    def testAddChildrenPackedHWithBorderEnabledTwoLayers(self):
        layout = fixtures.twoLayers((600, 100), border=1)
        self.assertEqual(2, len(layout.children))
        [self.assertEqual(1, border) for border in layout.borders]
        [
            self.addStuffToImage(child, 1, self.RandomColour())
            for child
            in layout.children
        ]
        [
            self.assertEqual((298, 98), child.size)
            for child
            in layout.children
        ]

    def testAddChildrenPackedHWithBorderEnabledThreeLayers(self):
        layout = fixtures.threeLayers((600, 100), border=1)
        [
            self.addStuffToImage(child, 1, self.RandomColour())
            for child
            in layout.children
        ]
        [
            self.assertEqual((198, 98), child.size)
            for child
            in layout.children
        ]
        self.assertEqual(0, layout._showSparePixels())

    def testAddChildrenPackedHWithBorderEnabledFourLayers(self):
        layout = fixtures.fourLayers((600, 100), border=1)
        [
            self.addStuffToImage(child, 1, self.RandomColour())
            for child
            in layout.children
        ]
        [
            self.assertEqual((148, 98), child.size)
            for child
            in layout.children
        ]
        self.assertEqual(0, layout._showSparePixels())

    def testAddChildrenPackedHWithBorderEnabledFiveLayers(self):
        layout = fixtures.fiveLayers((600, 100), border=1)
        [
            self.addStuffToImage(child, 1, self.RandomColour())
            for child
            in layout.children
        ]
        [
            self.assertEqual((118, 98), child.size)
            for child
            in layout.children
        ]
        self.assertEqual(0, layout._showSparePixels())

    def testAddChildrenPackedHWithBorderEnabledSixLayers(self):
        layout = fixtures.layoutWithLayers((600, 100), 6, border=1)
        [
            self.addStuffToImage(child, 1, self.RandomColour())
            for child
            in layout.children
        ]
        [
            self.assertEqual((98, 98), child.size)
            for child
            in layout.children
        ]
        self.assertEqual(0, layout._showSparePixels())

    def testAddChildrenPackedHWithBorderEnabledSevenLayers(self):
        layout = fixtures.layoutWithLayers((600, 100), 7, border=1)
        [
            self.addStuffToImage(child, 1, self.RandomColour())
            for child
            in layout.children
        ]
        [
            self.assertEqual((84, 98), child.size)
            for child
            in layout.children
        ]
        self.assertEqual(0, layout._showSparePixels())

    def testAddChildrenPackedHWithBorderEnabledEightLayers(self):
        layout = fixtures.layoutWithLayers((600, 100), 8, border=1)
        [
            self.addStuffToImage(child, 1, self.RandomColour())
            for child
            in layout.children
        ]
        _7LayerWidth = 73
        [
            self.assertEqual((_7LayerWidth, 98), child.size)
            for child
            in layout.children
        ]
        self.assertEqual(0, layout._showSparePixels())
        self.drawAndWrite(
                layout, "testChildrenPackedHWithBorderEnabledEightLayers.png")

    # REVISIT
    def testGenerating2Layers(self):
        layout = fixtures.threeLayers((400, 100), border=3)
        layout.imageMode = "RGB"
        self.addStuffToImage(layout, 0, self.RandomColour())
        [
            self.addStuffToImage(child, i, self.RandomColour())
            for i, child
            in enumerate(layout.children)
        ]

        layout.draw()
        self.assertEqual(0, layout._showSparePixels())

        self.drawAndWrite(layout, "testGenerating2Layers.png")
        slotsW = sum([slot[0] for slot in layout._slots])
        borderW = layout.borders[1] + layout.borders[3]
        spacersW = sum(layout._spacers)
        padding = sum(layout._paddings)

        self.assertEqual(6, borderW)
        self.assertEqual(7, spacersW)
        # FIXME - 2 spare things - should be 4
        self.assertEqual(1, padding)
        # FIXME - 2 spare things - should be 0
        self.assertEqual(0, layout.size[0] - slotsW - borderW - spacersW)
        [
            self.assertEqual((129, 94), child.size)
            for child
            in layout.children
        ]

    def testTreeOfLayers(self):
        l1 = fixtures.childlessLayout((400, 200), border=5)
        l2 = l1.addLayer(3)
        l3 = l2.addLayer()

        self.addStuffToImage(l1)
        self.addStuffToImage(l2)
        self.addStuffToImage(l3)

        self.drawAndWrite(l1, "testTreeOfLayers.png")
        self.assertEqual(l1.size, (400, 200))
        self.assertEqual(l2.size, (390, 190))
        self.assertEqual(l3.size, (384, 184))

    def test_calcDrawableSize(self):
        layout = fixtures.childlessLayout((400, 200), border=0)
        self.assertEqual((400, 200), layout._calcDrawableSize())

        layout = fixtures.childlessLayout((400, 200), border=1)
        self.assertEqual((398, 198), layout._calcDrawableSize())

        layout = fixtures.childlessLayout((400, 200), border=((1, 2), 1))
        self.assertEqual((396, 198), layout._calcDrawableSize())

        layout = fixtures.childlessLayout((400, 200), border=((1, 2, 3, 4), 1))
        self.assertEqual((394, 196), layout._calcDrawableSize())

    def test_calcSpacerErrorMarginBorder5OneChild(self):
        layout = fixtures.oneLayer((400, 200), border=5)
        self.assertEqual(0, layout._showSparePixels())

    def test_calcSpacerErrorMarginBorder5TwoChildren(self):
        layout = fixtures.twoLayers((400, 200), border=5)
        self.assertEqual(0, layout._showSparePixels())

    def test_calcSpacerErrorMarginBorder5ThreeChildren(self):
        layout = fixtures.threeLayers((400, 200), border=5)
        self.assertEqual(0, layout._showSparePixels())

    def test_calcSpacerErrorMarginBorder3(self):
        layout = Layout((40, 200), border=3)
        layout.addLayer()
        self.assertEqual(0, sum(layout._calcPaddings()))
        self.assertEqual(0, layout._showSparePixels())
        layout.addLayer()
        self.assertEqual(0, layout._showSparePixels())
        self.assertEqual(1, sum(layout._paddings))
        layout.addLayer()
        self.assertEqual(0, layout._showSparePixels())
        self.assertEqual(1, sum(layout._paddings))

    def test_spacersWidth(self):
        layout = Layout((400, 200), border=1)
        layout.addLayer()
        layout.addLayer()
        # VERIFIED correct unit test
        self.assertEqual(0, layout._showSparePixels())
        # VERIFIED correct unit test
        self.assertEqual(2, layout._sumSpacersWidth())

        layout.addLayer()
        self.assertEqual(2, layout._sumSpacersWidth())
        self.assertEqual(0, layout._showSparePixels())

    def test_calcSlotWidthBorder5(self):
        layout = Layout((400, 200), border=5)

        # No Layers added yet
        self.assertEqual((390, 190), layout._calcSlotSize())
        layout.addLayer()
        self.assertEqual((390, 190), layout._calcSlotSize())
        layout.addLayer()
        self.assertEqual((192, 190), layout._calcSlotSize())

    def test_calcSlotWidthBorder4(self):
        layout = Layout((400, 200), border=4)
        layout.addLayer()
        self.assertEqual((392, 192), layout._calcSlotSize())
        layout.addLayer()
        self.assertEqual((194, 192), layout._calcSlotSize())

    def NewImage(self, layout, mode='P', colour=randint(0, 2)):
        img = Image.new(mode, layout.size, colour)
        if mode == 'P':
            img.putpalette(Layout.DEFAULT_PALETTE)
        return img

    def RandomColour(self):
        return (randint(64, 255), randint(64, 255), randint(64, 255))

    def NewDraw(self, img):
        return ImageDraw.Draw(img)

    def addStuffToImage(self, layout, childId=0, bgcolour=(128, 128, 1)):

        img = self.NewImage(layout, "RGB", bgcolour)
        draw = self.NewDraw(img)
        text = "{id}\nchildId:{cid}\ndepth:{d}\ns:{size}\ntl:{tl}".format(
            id=layout._id,
            cid=childId, d=layout._depth, size=layout.size, tl=layout.topLeft)
        tsize = draw.textsize(text)
        if(tsize[1] > layout.size[1]):
            text = "{id}/{cid} d:{d}\ns:{size} tl:{tl}".format(
                id=layout._id, cid=childId, d=layout._depth,
                size=layout.size, tl=layout.topLeft)
        # draw.rectangle((0, 0), 1)
        draw.text((5, 5), text)
        layout.setImage(img)
        layout.draw()


if __name__ == '__main__':
    unittest.main()
